/* modified Bessel function of the first kind for complex arguments
   developed by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

double powi(double x,double n)
  {
  double p;
  p=1.;
  while(n>0)
    {
    p*=x;
    n--;
    }
  while(n<0)
    {
    p/=x;
    n++;
    }
  return(p);
  }

int BesselI(double x,double y,int n,double*re,double*im)
  {
  int i,ii,j,jj,k,l,m,ma,ml,mr,nb,nc,ne,nn,ns;
  double pi,p1i,p1r,p2i,p2r,pr,psi,psr,s,sp,si,sr,tai,tar,tbi,tbr,tci,tcr,test,t2,zi,zr;
  tar=sqrt(x*x+y*y);
  ma=(int)tar;
  if((n>0&&ma<=10000)&&(fabs(x)<=88.))
    {
    s=-1.;
    nc=n;
    if(powi(tar,4)<pow(0.1,16.6))
      {
      tar=1.;
      tai=0.;
      tcr=0.25*(x*x-y*y);
      tci=0.5*x*y;
      re[0]=1.-s*tcr;
      im[0]=-s*tci;
      if(n!=1)
        {
        for(nn=2;nn<=n;nn++)
          {
          tbr=(tar*x-tai*y)/(2*nn-2);
          tai=(tar*y+tai*x)/(2*nn-2);
          tar=tbr;
          tbr=nn;
          re[nn-1]=tar*(1.-s*tcr/tbr)+tai*tci/tbr;
          im[nn-1]=tai*(1.-s*tcr/tbr)-tar*tci/tbr;
          }
        }
      }
    else
      {
      nb=n-ma;
      nn=ma+1;
      if(fabs(x)<fabs(y))
        {
        zi=-1./(y+x*x/y);
        zr=-x*zi/y;
        }
      else
        {
        zr=1./(x+y*y/x);
        zi=-y*zr/x;
        }
      p1r=1.;
      p1i=0.;
      pr=s*(2*nn)*zr;
      pi=s*(2*nn)*zi;
      test=2.*pow(10.,16.6);
      m=0;
      if(nb>=3)
        {
        t2=pow(10.,38.-16.6);
        ns=ma+2;
        ne=n-1;
        for(nn=ns;nn<=ne;nn++)
          {
          p2r=p1r;
          p2i=p1i;
          p1r=pr;
          p1i=pi;
          pr=s*((2*nn)*(p1r*zr-p1i*zi)-p2r);
          pi=s*((2*nn)*(p1i*zr+p1r*zi)-p2i);
          sp=sqrt(powi(pr,2)+powi(pi,2));
          if(sp>t2)
            goto iter;
          }
        nn=ne;
        tbi=fmax(fabs(pr),fabs(pi));
        tbi*=sqrt(2.*pow(10.,16.6)*sqrt((powi(pr/tbi,2)+powi(pi/tbi,2))*(powi(p1r/tbi,2)+powi(p1i/tbi,2))));
        test=fmax(test,tbi);
        goto series;
      iter:
        ns=nn+1;
        pr/=t2;
        pi/=t2;
        p1r/=t2;
        p1i/=t2;
        psr=pr;
        psi=pi;
        tcr=p1r;
        tci=p1i;
        test=pow(10.,16.6+16.6);
        while(1)
          {
          nn+=1;
          p2r=p1r;
          p2i=p1i;
          p1r=pr;
          p1i=pi;
          pr=s*((2*nn)*(p1r*zr-p1i*zi)-p2r);
          pi=s*((2*nn)*(p1i*zr+p1r*zi)-p2i);
          if(powi(pr,2)+powi(pi,2)>test)
            break;
          }
        tbr=sqrt((powi(p1r,2)+powi(p1i,2))/(powi(p2r,2)+powi(p2i,2)));
        tbi=nn/tar;
        if(tbr+1./tbr>2.*tbi)
          tbr=tbi+sqrt(powi(tbi,2)-1.);
        test=0.5*(1.-1./powi(tbr,2))/pow(10.,16.6);
        test=((powi(p1r,2)+powi(p1i,2))*test)*((powi(p2r,2)+powi(p2i,2))*test);
        pr=p1r*t2;
        pi=p1i*t2;
        nn-=1;
        ne=min(n,nn);
        for(nc=ns;nc<=ne;nc++)
          {
          p2r=tcr;
          p2i=tci;
          tcr=psr;
          tci=psi;
          psr=s*((2*nn)*(tcr*zr-tci*zi)-p2r);
          psi=s*((2*nn)*(tci*zr+tcr*zi)-p2i);
          if((powi(psr,2)+powi(psi,2))*(powi(tcr,2)+powi(tci,2))-test>0)
            goto skip;
          }
        nc=ne+1;
      skip:
        nc-=1;
        goto asymptotic;
        }
    series:
      while(1)
        {
        nn+=1;
        p2r=p1r;
        p2i=p1i;
        p1r=pr;
        p1i=pi;
        pr=s*((2*nn)*(p1r*zr-p1i*zi)-p2r);
        pi=s*((2*nn)*(p1i*zr+p1r*zi)-p2i);
        if(powi(pr/test,2)+powi(pi/test,2)>=1.)
          {
          if(m==1)
            break;
          m=1;
          tbi=fmax(fabs(pr),fabs(pi));
          tbr=sqrt((powi(pr/tbi,2)+powi(pi/tbi,2))/(powi(p1r/tbi,2)+powi(p1i/tbi,2)));
          tbi=(nn+1)/tar;
          if(tbr+1./tbr>2.*tbi)
            tbr=tbi+sqrt(powi(tbi,2)-1.);
          test/=sqrt(tbr-1.E0/tbr);
          if(powi(pr/test,2)+powi(pi/test,2)-1.>=0)
            break;
          }
        }
    asymptotic:
      nn+=1;
      tbr=x;
      i=0;
      if(tbr!=0)
        i=(int)((float)(1.1*tbr/fabs(tbr)));
      mr=4*((3+i)/2)-3-2*(1+i);
      k=(2*i+1)*i+1;
      l=nn-4*(nn/4);
      ml=2+8*((k*l)/4)-4*((k*l)/2);
      if(i==0&&(l==1||l==3))
        ml=0;
      l+=3-4*((l+3)/4);
      m=2+8*((k*l)/4)-4*((k*l)/2);
      if(i==0&&(l==1||l==3))
        m=0;
      jj=0;
      j=0;
      tbr=tbi=0.;
      if(fabs(pi)>fabs(pr))
        {
        tai=-1./(pi+pr*(pr/pi));
        tar=-(pr*tai)/pi;
        }
      else
        {
        tar=1./(pr+pi*(pi/pr));
        tai=-(pi*tar)/pr;
        }
      if(j==0)
        {
        sr=m*tar;
        si=m*tai;
        }
      else
        {
        sr=-m*tai;
        si=m*tar;
        }
      ne=nn-n;
      if(ne<0)
        {
        re[nn-1]=tar;
        im[nn-1]=tai;
        ne=-ne;
        for(l=1;l<=ne;l++)
          {
          ii=nn+l;
          re[ii-1]=im[ii-1]=0.;
          }
        }
      else
        {
        if(ne!=0)
          {
          for(l=1;l<=ne;l++)
            {
            nn-=1;
            tcr=tbr;
            tci=tbi;
            tbr=tar;
            tbi=tai;
            pr=2*nn*zr;
            pi=2*nn*zi;
            tar=pr*tbr-pi*tbi-s*tcr;
            tai=pr*tbi+pi*tbr-s*tci;
            j=(1-j)*jj;
            k=ml;
            ml=m;
            m=k*mr;
            if(j==0)
              {
              sr+=m*tar;
              si+=m*tai;
              }
            else
              {
              sr+=-m*tai;
              si+=m*tar;
              }
            }
          }
        re[nn-1]=tar;
        im[nn-1]=tai;
        if(nn>1)
          {
          nn-=1;
          pr=2*nn*zr;
          pi=2*nn*zi;
          re[nn-1]=pr*tar-pi*tai-s*tbr;
          im[nn-1]=pr*tai+pi*tar-s*tbi;
          if(nn==1)
            goto L_140;
          j=(1-j)*jj;
          k=ml;
          ml=m;
          m=k*mr;
          if(j==0)
            {
            sr+=m*re[nn-1];
            si+=m*im[nn-1];
            }
          else
            {
            sr+=-m*im[nn-1];
            si+=m*re[nn-1];
            }
          }
        else
          {
          sr-=tar;
          si-=tai;
          goto L_150;
          }
        }
      ne=nn-2;
      if(ne!=0)
        {
        for(l=1;l<=ne;l++)
          {
          nn-=1;
          pr=2*nn*zr;
          pi=2*nn*zi;
          re[nn-1]=pr*re[nn]-pi*im[nn]-s*re[nn+1];
          im[nn-1]=pr*im[nn]+pi*re[nn]-s*im[nn+1];
          j=(1-j)*jj;
          k=ml;
          ml=m;
          m=k*mr;
          if(j==0)
            {
            sr+=m*re[nn-1];
            si+=m*im[nn-1];
            }
          else
            {
            sr+=-m*im[nn-1];
            si+=m*re[nn-1];
            }
          }
        }
      re[0]=2.*(re[1]*zr-im[1]*zi)-s*re[2];
      im[0]=2.*(re[1]*zi+im[1]*zr)-s*im[2];
  L_140:
      sr+=re[0];
      si+=im[0];
  L_150:
      tcr=i*x;
      tci=i*y;
      tcr=exp(tcr);
      tbr=cos(tci);
      tbi=sin(tci);
      if(fabs(sr)<fabs(si))
        {
        tci=sr/si;
        tcr=(tcr/si)/(1.+tci*tci);
        tar=tcr*(tbr*tci+tbi);
        tai=tcr*(tbi*tci-tbr);
        }
      else
        {
        tci=si/sr;
        tcr=(tcr/sr)/(1.+tci*tci);
        tar=tcr*(tbr+tbi*tci);
        tai=tcr*(tbi-tbr*tci);
        }
      for(nn=1;nn<=n;nn++)
        {
        tbr=re[nn-1]*tar-im[nn-1]*tai;
        im[nn-1]=re[nn-1]*tai+im[nn-1]*tar;
        re[nn-1]=tbr;
        }
      }
    if(n==nc)
      return(0);
    }
  return(__LINE__);
  }

#define L 6
int l=0;
double Re[L],Im[L],Sm,Sx;

void CreateTB2(char*fname,int which)
  {
  int i,j,m=64,n=64;
  double x,y;
  FILE*fp;
  printf("creating: %s\n",fname);
  fp=fopen(fname,"wt");
  fprintf(fp,"%i\n",n);
  for(i=0;i<n;i++)
    {
    x=i*6./(n-1);
    fprintf(fp,"%lG\n",x);
    }
  fprintf(fp,"%i\n",m);
  for(j=0;j<m;j++)
    {
    y=-3.+j*6./(m-1);
    fprintf(fp,"%lG\n",y);
    }
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  fprintf(fp,"%i\n",n*m);
  for(j=0;j<m;j++)
    {
    y=-3.+j*6./(m-1);
    for(i=0;i<n;i++)
      {
      x=i*6./(n-1);
      BesselI(x,y,L,Re,Im);
      if(which%2)
        {
        fprintf(fp,"%lG\n",Re[l]);
        if(Re[l]<Sm)
          Sm=Re[l];
        if(Re[l]>Sx)
          Sx=Re[l];
        }
      else
        {
        fprintf(fp,"%lG\n",Im[l]);
        if(Im[l]<Sm)
          Sm=Im[l];
        if(Im[l]>Sx)
          Sx=Im[l];
        }
      }
    }
  fclose(fp);
  }

void CreateTP2(char*nam1,char*nam2,char*nam3)
  {
  FILE*fp;
  printf("creating file: %s\n",nam3);
  fp=fopen(nam3,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!ALTER\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"NOLEGEND 126\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
  fprintf(fp,"I(%i,z)\n",l);
  fprintf(fp,"X\n");
  fprintf(fp,"0 6\n");
  fprintf(fp,"Y\n");
  fprintf(fp,"-3 3\n");
  fprintf(fp,"%s\n",nam1);
  fprintf(fp,"NEWFILE 2\n");
  fprintf(fp,"%s\n",nam2);
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  char buf1[32],buf2[32],buf3[32];
  if(argc==2)
    {
    l=atoi(argv[1]);
    if(l<0)
      {
      printf("l<0\n");
      return(0);
      }
    if(l>=L)
      {
      printf("l>=%i\n",L);
      return(0);
      }
    }
  sprintf(buf1,"I%ir.tb2",l);
  sprintf(buf2,"I%ii.tb2",l);
  sprintf(buf3,"I%i.tp2",l);
  CreateTB2(buf1,1);
  CreateTB2(buf2,2);
  CreateTP2(buf1,buf2,buf3);
  return(0);
  }
