/* various Riemann surfaces
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "../include/complex.hpp"
#define WIN32_LEAN_AND_MEAN
#include <windows.h>

DWORD Palette[]={
  0x000000,0x9C9C9C,0x0000FC,0x0008FC,0x0010FC,0x0018FC,0x0020FC,0x0028FC,
  0x0030FC,0x0038FC,0x0040FC,0x0048FC,0x0050FC,0x0058FC,0x0060FC,0x0068FC,
  0x0070FC,0x0078FC,0x0080FC,0x0088FC,0x0090FC,0x0098FC,0x00A0FC,0x00A8FC,
  0x00B0FC,0x00B8FC,0x00C0FC,0x00C8FC,0x00D0FC,0x00D8FC,0x00E0FC,0x00E8FC,
  0x00F0FC,0x00FCFC,0x00FCF0,0x00FCE8,0x00FCE0,0x00FCD8,0x00FCD0,0x00FCC8,
  0x00FCC0,0x00FCB8,0x00FCB0,0x00FCA8,0x00FCA0,0x00FC98,0x00FC90,0x00FC88,
  0x00FC80,0x00FC78,0x00FC70,0x00FC68,0x00FC60,0x00FC58,0x00FC50,0x00FC48,
  0x00FC40,0x00FC38,0x00FC30,0x00FC28,0x00FC20,0x00FC18,0x00FC10,0x00FC08,
  0x00FC00,0x08FC00,0x10FC00,0x18FC00,0x20FC00,0x28FC00,0x30FC00,0x38FC00,
  0x40FC00,0x48FC00,0x50FC00,0x58FC00,0x60FC00,0x68FC00,0x70FC00,0x78FC00,
  0x80FC00,0x88FC00,0x90FC00,0x98FC00,0xA0FC00,0xA8FC00,0xB0FC00,0xB8FC00,
  0xC0FC00,0xC8FC00,0xD0FC00,0xD8FC00,0xE0FC00,0xE8FC00,0xF0FC00,0xFCFC00,
  0xFCF000,0xFCE800,0xFCE000,0xFCD800,0xFCD000,0xFCC800,0xFCC000,0xFCB800,
  0xFCB000,0xFCA800,0xFCA000,0xFC9800,0xFC9000,0xFC8800,0xFC8000,0xFC7800,
  0xFC7000,0xFC6800,0xFC6000,0xFC5800,0xFC5000,0xFC4800,0xFC4000,0xFC3800,
  0xFC3000,0xFC2800,0xFC2000,0xFC1800,0xFC1000,0xFC0800,0xFC0000,0xFCFCFC,
  0x040404,0x0C0C0C,0x141414,0x1C1C1C,0x242424,0x2C2C2C,0x343434,0x3C3C3C,
  0x444444,0x4C4C4C,0x545454,0x5C5C5C,0x646464,0x6C6C6C,0x747474,0x7C7C7C,
  0x848484,0x8C8C8C,0x949494,0xA4A4A4,0xACACAC,0xB4B4B4,0xBCBCBC,0xC4C4C4,
  0xCCCCCC,0xD4D4D4,0xDCDCDC,0xE4E4E4,0xECECEC,0xF4F4F4,0x0000A0,0x00A0A0,
  0x00A000,0xD0D000,0xE07040,0xA00000,0xA00060,0xFC00A0,0x6060A0,0xA0A0FC,
  0x7F3F00,0x00D200,0x7F00FF,0x2D2DFF,0xFFBF7F,0xD20000,0xFF007F,0xFF00FF};

BYTE        black   =0x00;
BYTE   dark_blue    =0x9E;
BYTE  royal_blue    =0xAB;
BYTE        blue    =0x02;
BYTE        brown   =0xA8;
BYTE   dark_cyan    =0x9F;
BYTE        cyan    =0x21;
BYTE   dark_gray    =0x83;
BYTE medium_gray    =0x8F;
BYTE        gray    =0x01;
BYTE  light_gray    =0x99;
BYTE   dark_green   =0xA0;
BYTE medium_green   =0xA9;
BYTE        green   =0x40;
BYTE        lavender=0xAA;
BYTE   dark_magenta =0xA4;
BYTE medium_magenta =0xA5;
BYTE  light_magenta =0xAF;
BYTE   dark_orange  =0xA2;
BYTE        orange  =0x6F;
BYTE        pink    =0xAC;
BYTE   dark_purple  =0xA6;
BYTE  light_purple  =0xA7;
BYTE   dark_red     =0xA3;
BYTE medium_red     =0xAD;
BYTE        red     =0x7E;
BYTE bright_red     =0xAE;
BYTE        white   =0x7F;
BYTE   dark_yellow  =0xA1;
BYTE        yellow  =0x5F;

void Abort(int line,char*format,...)
  {
  va_list arg_marker;
  va_start(arg_marker,format);
  fprintf(stderr,"\nerror %i: ",line);
  vfprintf(stderr,format,arg_marker);
  fprintf(stderr,"\n");
  _fcloseall();
  exit(1);
  }

BYTE rainbow(double x,double lo,double hi)
  {
  int i;
  i=(int)((int)blue+(x-lo)*((int)red-(int)blue)/(hi-lo));
  if(i<=(int)blue)
    return(blue);
  if(i>=(int)red)
    return(red);
  return((BYTE)(unsigned)i);
  }

void Surface(int nx,int ny,char*fname)
  {
  int x,y;
  double a;
  complex g,z;
  FILE*fp;
  printf("creating surface: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"%i nx\n",nx);
  for(x=0;x<nx;x++)
    {
    z.re=-2.+x*4./(nx-1);
    fprintf(fp,"%lG\n",z.re);
    }
  fprintf(fp,"%i ny\n",ny);
  for(y=0;y<ny;y++)
    {
    z.im=-2.+y*4./(ny-1);
    fprintf(fp,"%lG\n",z.im);
    }
  fprintf(fp,"%i nz=nx*ny\n",nx*ny);
  for(y=0;y<ny;y++)
    {
    z.im=-2.+y*4./(ny-1);
    for(x=0;x<nx;x++)
      {
      z.re=-4.+x*8./(nx-1);
#ifdef GAMMA
      g=gamma(z);
#else
      g=cos(z);
#endif
      a=hypot(g.re,g.im);
      if(a>6.)
        a=6.;
      fprintf(fp,"%lG\n",a);
      }
    }
  fclose(fp);
  }

void Body(int nu,int nv,char*fname)
  {
  int i,j;
  double u,v,x,y;
  complex w,z;
  FILE*fp;
  printf("creating 3D body: %s\n",fname);
  if((fp=fopen(fname,"wt"))==NULL)
    Abort(__LINE__,"can't create output file");
  fprintf(fp,"%i nodes\n",nu*nv);
  for(i=0;i<nu;i++)
    {
    u=-6.+i*7./(nu-1);
    w.re=u;
    for(j=0;j<nv;j++)
      {
      v=-9.+j*18./(nv-1);
      w.im=v;
      z=exp(w);
      x=z.re;
      y=z.im;
      fprintf(fp,"%lG,%lG,%lG\n",x,y,v);
      }
    }
  fprintf(fp,"%i elements\n",(nu-1)*(nv-1));
  for(i=1;i<nu;i++)
    {
    u=-6.+i*7./(nu-1);
    for(j=1;j<nv;j++)
      {
      v=-9.+j*18./(nv-1);
      fprintf(fp,"%i %i %i %i RGB=0x%06X\n",(i-1)*nv+j,(i-1)*nv+j+1,i*nv+j+1,i*nv+j,Palette[rainbow(v,-9.,9.)]);
      }
    }
  fclose(fp);
  }

int randbetween(int lo,int hi)
  {
  return(lo+rand()%(hi-lo+1));
  }

int main(int argc,char**argv,char**envp)
  {
#ifdef CREATE_BODY
  complex w,x,y,z;
  Body(50,50,"rsurf.3dv");
#elif RANDOM_POINTS
  int i;
  for(i=0;i<10;i++)
    {
    x.re=randbetween(-30,30)/10.;
    x.im=randbetween(-30,30)/10.;
    y.re=randbetween(-30,30)/10.;
    y.im=randbetween(-30,30)/10.;
    z=x/y;
    w=exp(log(x)-log(y));
    printf("%lG:%lG ? %lG:%lG\n",z.re,z.im,w.re,w.im);
    }
#else
  Surface(100,100,"rsurf.tb2");
#endif
  return(0);
  }
