/* potential flow through a natural draft cooling tower
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#define _USE_MATH_DEFINES
#include <float.h>

#include <math.h>

double Ri=150; /* radius at inlet */
double Rt= 90; /* radius at throat */
double Re=100; /* radius at exit */
double Hi= 40; /* height of inlet */
double He=450; /* height at exit */
double Sm,Sx;

double Radius(double Z)
  {
  double E1,E2,E3,Ht;
  Z=fabs(Z);
  if(Z<=Hi)
    return(Ri);
  if(Z>=He)
    return(Re);
  E1=Re/Rt;
  E1=sqrt(E1*E1-1.);
  E2=Ri/Rt;
  E2=sqrt(E2*E2-1.);
  Ht=(E1*Hi+E2*He)/(E1+E2);
  E3=Rt*E1/(He-Ht);
  return(hypot(E3*(Z-Ht),Rt));
  }

double Psi(double a,double b,double r,double z)
  {
  double p;
  if(z>b)
    p=b/z;
  else if(z<-b)
    p=-b/z;
  else
    p=1;
  p*=sqrt(2);
  if(fabs(r)<=a)
    return(p*(3-r*r/a/a)*r/a/2);
  else
    return(p*a*a/r/fabs(r));
  }

double Stream(double r,double z)
  {
  double a,b,c;
  a=Radius(z);
  b=(He-Hi)/2;
  c=(He+Hi)/2;
  return(Psi(a,b,r,z-c)-Psi(a,b,r,z+c));
  }

double U1(double r,double z)
  {
  double dz=He/1E4;
  return((Stream(r,z-dz/2)-Stream(r,z+dz/2))/dz);
  }

double V1(double r,double z)
  {
  double dr=Ri/1E4;
  return(((r+dr/2)*Stream(r+dr/2,z)-(r-dr/2)*Stream(r-dr/2,z))/dr/r);
  }

void CreateTB2(char*name,int Nr,int Nz)
  {
  int r,z;
  double R,S,Z;
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"%li Nx\n",Nr);
  for(r=0;r<Nr;r++)
    {
    R=(2*r-Nr+1)*4*Ri/(Nr-1);
    fprintf(fp,"%lG\n",R);
    }
  fprintf(fp,"%li Ny\n",Nz);
  for(z=0;z<Nz;z++)
    {
    Z=(2*z-Nz+1)*2*He/(Nz-1);
    fprintf(fp,"%lG\n",Z);
    }
  Sm=DBL_MAX;
  Sx=-DBL_MAX;
  fprintf(fp,"%li Nz\n",Nr*Nz);
  for(z=0;z<Nz;z++)
    {
    Z=(2*z-Nz+1)*2*He/(Nz-1);
    for(r=0;r<Nr;r++)
      {
      R=(2*r-Nr+1)*4*Ri/(Nr-1);
      S=Stream(R,Z);
      Sm=min(Sm,S);
      Sx=max(Sx,S);
      fprintf(fp,"%lG\n",S);
      }
    }
  fclose(fp);
  }

void CreateV2D(char*name,int Nr,int Nz)
  {
  char cbuf[128];
  int r,z;
  double R,Z;
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  for(z=0;z<Nz;z++)
    {
    Z=(2*z-Nz+1)*2*He/(Nz-1);
    for(r=0;r<Nr;r++)
      {
      R=(2*r-Nr+1)*4*Ri/(Nr-1);
      if(fabs(R)<1.)
        continue;
      sprintf(cbuf,"%lG %lG %lG %lG\n",R,Z,U1(R,Z),V1(R,Z));
      if(!strstr(cbuf,"NAN"))
        fputs(cbuf,fp);
      }
    }
  fclose(fp);
  }

void CreateTP2(char*name)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm,Sx);
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
  fprintf(fp,"Potential Flow through a Natural Draft Cooling Tower\n");
  fprintf(fp,"Radius\n");
  fprintf(fp,"%lG %lG\n",-2.*Ri,2.*Ri);
  fprintf(fp,"Height\n");
  fprintf(fp,"0 %lG\n",2.*He);
  fprintf(fp,"cooling_tower.tb2\n");
  fprintf(fp,"ARROW\n");
  fprintf(fp,"0 0\n");
  fprintf(fp,"NEWFILE\n");
  fprintf(fp,"cooling_tower.v2d\n");
  fprintf(fp,"NOCOLOR\n");
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
  CreateTB2("cooling_tower.tb2",99,99);
  CreateV2D("cooling_tower.v2d",99,99);
  CreateTP2("cooling_tower.tp2");
  exit(0);
  }
