/* Gauss error function for complex arguments
   developed by Dudley J. Benton
   dudley.benton@gmail.com */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#include "../include/complex.hpp"

complex erf(complex z)
  {
  int i;
  double c,c2,dm,ef1,ef2,pm,qf1,qf2,qm,r,sm1,sm2,sn,ss,sz1,sz2,tm1,tm2,ts1,ts2,x,y;
  static double cd[18]={0.00000000000000E00,2.08605856013476E-2,
    8.29806940495687E-2,1.85421653326079E-1,3.27963479382361E-1,
    5.12675279912828E-1,7.45412958045105E-1,1.03695067418297E00,
    1.40378061255437E00,1.86891662214001E00,2.46314830523929E00,
    3.22719383737352E00,4.21534348280013E00,5.50178873151549E00,
    7.19258966683102E00,9.45170208076408E00,1.25710718314784E01,
    1.72483537216334E01};
  static double ce[18]={8.15723083324096E-2,1.59285285253437E-1,
    1.48581625614499E-1,1.33219670836245E-1,1.15690392878957E-1,
    9.78580959447535E-2,8.05908834297624E-2,6.40204538609872E-2,
    4.81445242767885E-2,3.33540658473295E-2,2.05548099470193E-2,
    1.07847403887506E-2,4.55634892214219E-3,1.43984458138925E-3,
    3.07056139834171E-4,3.78156541168541E-5,2.05173509616121E-6,
    2.63564823682747E-8};
  c=1./sqrt(M_PI);
  x=z.re;
  y=z.im;
  sn=1.;
  if(x<0.)
    {
    x=-x;
    y=-y;
    sn=-1.;
    }
  r=x*x+y*y;
  sz1=x*x-y*y;
  sz2=2.*x*y;
  if(r>1.)
    {
    if(r>=38.)
      {
      qf1=sz1/(r*r);
      qf2=-sz2/(r*r);
      qm=c*exp(-sz1);
      ts1=qm*cos(-sz2);
      ts2=qm*sin(-sz2);
      tm1=(ts1*x+ts2*y)/r;
      tm2=-(ts1*y-ts2*x)/r;
      sm1=tm1;
      sm2=tm2;
      pm=-0.5;
      while(1)
        {
        pm+=1.;
        ts1=tm1*qf1-tm2*qf2;
        ts2=tm1*qf2+tm2*qf1;
        tm1=-pm*ts1;
        tm2=-pm*ts2;
        if(fabs(sm1)+fabs(tm1)==fabs(sm1))
          if(fabs(sm2)+fabs(tm2)==fabs(sm2))
            break;
        sm1+=tm1;
        sm2+=tm2;
        if(pm>=25.5)
          break;
        }
      if(x>=0.01)
        {
        ef1=sm1;
        ef2=sm2;
        }
      else
        {
        sn=-sn;
        return(complex(sn*sm1,sn*sm2));
        }
      }
    else if(sz1+0.064*sz2*sz2>0.)
      {
      sm1=sm2=0.;
      qm=c*exp(-sz1);
      ts1=qm*cos(-sz2);
      ts2=qm*sin(-sz2);
      qf1=ts1*x-ts2*y;
      qf2=ts1*y+ts2*x;
      for(i=1;i<=18;i++)
        {
        ts1=sz1+cd[i-1];
        ts2=sz2;
        ss=ts1*ts1+ts2*ts2;
        tm1=ce[i-1]*ts1/ss;
        tm2=-ce[i-1]*ts2/ss;
        sm1+=tm1;
        sm2+=tm2;
        }
      ef1=qf1*sm1-qf2*sm2;
      ef2=qf1*sm2+qf2*sm1;
      }
    else
      goto TaylorSeries;
    ef1=sn*(1.-ef1);
    ef2*=-sn;
    return(complex(ef1,ef2));
    }
TaylorSeries:
  c2=c+c;
  tm1=c2*x;
  tm2=c2*y;
  sm1=tm1;
  sm2=tm2;
  pm=0.;
  while(1)
    {
    pm+=1.;
    dm=2.*pm+1.;
    ts1=tm1*sz1-tm2*sz2;
    ts2=tm1*sz2+tm2*sz1;
    tm1=-ts1/pm;
    tm2=-ts2/pm;
    ts1=tm1/dm;
    ts2=tm2/dm;
    if(fabs(sm1)+fabs(ts1)==fabs(sm1))
      if(fabs(sm2)+fabs(ts2)==fabs(sm2))
        break;
    sm1+=ts1;
    sm2+=ts2;
    }
  return(complex(sn*sm1,sn*sm2));
  }

complex Sm,Sx;

void CreateTB2(char*fname,int which)
  {
  int i,j,m=64,n=64;
  complex w,z;
  FILE*fp;
  printf("transformation: %s\n",fname);
  fp=fopen(fname,"wt");
  fprintf(fp,"%i\n",n);
  for(i=0;i<n;i++)
    {
    z.re=-3.+i*6./(n-1);
    fprintf(fp,"%lG\n",z.re);
    }
  fprintf(fp,"%i\n",m);
  for(j=0;j<m;j++)
    {
    z.im=-3.+j*6./(m-1);
    fprintf(fp,"%lG\n",z.im);
    }
  if(which%2)
    {
    Sm.re=DBL_MAX;
    Sx.re=-DBL_MAX;
    }
  else
    {
    Sm.im=DBL_MAX;
    Sx.im=-DBL_MAX;
    }
  fprintf(fp,"%i\n",n*m);
  for(j=0;j<m;j++)
    {
    z.im=-3.+j*6./(m-1);
    for(i=0;i<n;i++)
      {
      z.re=-3.+i*6./(n-1);
      w=erf(z);
      if(w.re>4.)
        w.re=4.;
      else if(w.re<-4.)
        w.re=-4.;
      if(w.im>4.)
        w.im=4.;
      else if(w.im<-4.)
        w.im=-4.;
      if(which%2)
        {
        fprintf(fp,"%lG\n",w.re);
        if(w.re<Sm.re)
          Sm.re=w.re;
        if(w.re>Sx.re)
          Sx.re=w.re;
        }
      else
        {
        fprintf(fp,"%lG\n",w.im);
        if(w.im<Sm.im)
          Sm.im=w.im;
        if(w.im>Sx.im)
          Sx.im=w.im;
        }
      }
    }
  fclose(fp);
  }

void CreateTP2(char*name)
  {
  FILE*fp;
  printf("creating file: %s\n",name);
  fp=fopen(name,"wt");
  fprintf(fp,"TP2\n");
  fprintf(fp,"!ALTER\n");
  fprintf(fp,"!PLOT\n");
  fprintf(fp,"EQUAL\n");
  fprintf(fp,"TICKS\n");
  fprintf(fp,"LEGENDS\n");
  fprintf(fp,"-1 5 2\n");
  fprintf(fp,"RAINBOW\n");
  fprintf(fp,"CONTOUR\n");
  fprintf(fp,"%lG 0 %lG\n",Sm.re,Sx.re);
  fprintf(fp,"NOLEGEND 126\n");
  fprintf(fp,"ZTABLE\n");
  fprintf(fp,"CENTER\n");
  fprintf(fp,"ABOVE\n");
  fprintf(fp,"w=erf(z)\n");
  fprintf(fp,"X\n");
  fprintf(fp,"-3 3\n");
  fprintf(fp,"Y\n");
  fprintf(fp,"-3 3\n");
  fprintf(fp,"erf1.tb2\n");
  fprintf(fp,"NEWFILE 2\n");
  fprintf(fp,"erf2.tb2\n");
  fprintf(fp,"!END\n");
  fclose(fp);
  }

int main(int argc,char**argv,char**envp)
  {
#ifdef csv
  int i,j,m=101,n=11;
  double x,y;
  complex w,z;
  for(j=0;j<m;j++)
    {
    x=-3.+j*6./(m-1);
    printf("%lG",x);
    for(i=0;i<n;i++)
      {
      y=-3.+i*6./(n-1);
      z=complex(x,y);
      w=erf(z);
      printf(",%lG,%lG",w.re,w.im);
      }
    printf("\n");
    }
#else
  CreateTB2("erf1.tb2",1);
  CreateTB2("erf2.tb2",2);
  CreateTP2("erf.tp2");
#endif
  return(0);
  }
