#ifndef __COMPLEX_H
#define __COMPLEX_H 1
#ifndef __MATH_H
#include <math.h>
#endif

unsigned int nan2i[2]={0x00000000,0x7FF80000};
#define __nan (*(double*)&nan2i)

class complex
  {
 public:
   friend double real (complex&);
   friend double imag (complex&);
   friend complex cos (complex&);
   friend complex cosh (complex&);
   friend complex sin (complex&);
   friend complex sinh (complex&);
   friend complex tan (complex&);
   friend complex tanh (complex&);
   friend complex log (complex&);
   friend complex log10 (complex&);
   friend complex sqrt (complex&);
   friend double abs (complex&);
   friend complex conj (complex&);
   friend double norm (complex&);
   friend double modulus (complex&);
   friend double arg (complex&);
   friend complex asin (complex&);
   friend complex acos (complex&);
   friend complex atan (complex&);
   friend complex asinh (complex&);
   friend complex atanh (complex&);
   friend complex exp (complex&);
   friend complex polar (double r,double theta = 0);
   friend complex operator + (complex&,complex&);
   friend complex operator + (double,complex&);
   friend complex operator + (complex&,double);
   friend complex operator - (complex&,complex&);
   friend complex operator - (double,complex&);
   friend complex operator - (complex&,double);
   friend complex operator * (complex&,complex&);
   friend complex operator * (double,complex&);
   friend complex operator * (complex&,double);
   friend complex operator / (complex&,complex&);
   friend complex operator / (double,complex&);
   friend complex operator / (complex&,double);
   friend complex pow(complex& lhs,double rhs);
   friend complex pow(complex& lhs,complex & rhs);
   friend complex pow(double lhs,complex & rhs);
   friend complex pow(complex& lhs,int rhs);
   friend int operator && (complex&,complex&);
   friend int operator && (double,complex&);
   friend int operator && (complex&,double);
   friend int operator || (complex&,complex&);
   friend int operator || (double,complex&);
   friend int operator || (complex&,double);
   friend int operator != (complex&,complex&);
   friend int operator != (double,complex&);
   friend int operator != (complex&,double);
   friend int operator == (complex&,complex&);
   friend int operator == (double,complex&);
   friend int operator == (complex&,double);
 public:
   complex ();
   complex (double r,double i = 0);
   complex (complex &);
   double& real ();
   double& imag ();
   complex& operator = (complex&);
   complex& operator = (double);
   complex& operator += (complex&);
   complex& operator += (double);
   complex& operator -= (complex&);
   complex& operator -= (double);
   complex& operator *= (complex&);
   complex& operator *= (double);
   complex& operator /= (complex&);
   complex& operator /= (double);
   int operator ! () const;
   complex operator - () const;
   double re;
   double im;
   };

complex I=complex(0.,1.);

inline complex::complex()
  {
  re=im=__nan;
  }

inline complex::complex(double r,double i): re(r),im(i) { } inline complex::complex(complex& z)
  {
  re=z.re;
  im=z.im;
  }

inline double real(complex& z) {return z.re;}

inline double imag(complex& z) {return z.im;}

inline double& complex::real() {return re;}

inline double& complex::imag() {return im;}

inline complex operator +(complex& lhs,complex& rhs)
  {
  return complex(lhs.re+rhs.re,lhs.im+rhs.im);
  }

inline complex operator +(double lhs,complex& rhs)
  {
  return complex(lhs+rhs.re,rhs.im);
  }

inline complex operator +(complex& lhs,double rhs)
  {
  return complex(lhs.re+rhs,lhs.im);
  }

inline complex operator -(complex& lhs,complex& rhs)
  {
  return complex(lhs.re-rhs.re,lhs.im-rhs.im);
  }

inline complex operator -(double lhs,complex& rhs)
  {
  return complex(lhs-rhs.re,-rhs.im);
  }

inline complex operator -(complex& lhs,double rhs)
  {
  return complex(lhs.re-rhs,lhs.im);
  }

inline complex operator *(complex& lhs,complex& rhs)
  {
  return complex(lhs.re*rhs.re-lhs.im*rhs.im,lhs.re*rhs.im+lhs.im*rhs.re);
  }

inline complex operator *(double lhs,complex& rhs)
  {
  return complex(lhs*rhs.re,lhs*rhs.im);
  }

inline complex operator *(complex& lhs,double rhs)
  {
  return complex(lhs.re*rhs,lhs.im*rhs);
  }

inline complex operator /(complex& lhs,double rhs)
  {
  return complex(lhs.re/rhs,lhs.im /rhs);
  }

inline complex operator /(double lhs,complex& rhs)
  {
  double a;
  a=rhs.re*rhs.re+rhs.im*rhs.im;
  return complex(lhs*rhs.re/a,-lhs*rhs.im/a);
  }

inline complex operator /(complex& lhs,complex& rhs)
  {
  double a;
  a=rhs.re*rhs.re+rhs.im*rhs.im;
  return complex(
    (((lhs.re)*(rhs.re))+((lhs.im)*(rhs.im)))/a,
    (((rhs.re)*(lhs.im))-((lhs.re)*(rhs.im)))/a);
  }

inline complex complex::operator -() const
  {
  return complex(-re,-im);
  }

inline int complex::operator !() const
  {
  return(re==0)&&(im==0);
  }

inline complex & complex::operator =(complex& z)
  {
  re=z.re;
  im=z.im;
  return *this;
  }

inline complex & complex::operator =(double x)
  {
  re=x;
  im=0;
  return *this;
  }

inline complex & complex::operator *=(complex& z)
  {
  *this=*this*z;
  return *this;
  }

inline complex & complex::operator *=(double x)
  {
  re*=x;
  im*=x;
  return *this;
  }

inline complex & complex::operator /=(complex& z)
  {
  *this=*this/z;
  return *this;
  }

inline complex & complex::operator /=(double x)
  {
  re/=x;
  im/=x;
  return *this;
  }

inline complex & complex::operator +=(complex& z)
  {
  re+=z.re;
  im+=z.im;
  return *this;
  }

inline complex & complex::operator +=(double x)
  {
  re+=x;
  return *this;
  }

inline complex & complex::operator -=(complex& z)
  {
  re-=z.re;
  im-=z.im;
  return *this;
  }

inline complex & complex::operator -=(double x)
  {
  re-=x;
  return *this;
  }

inline int operator&&(complex & lhs,complex & rhs)
  {
  return((lhs.re!=0)||(lhs.im!=0))&&((rhs.re!=0)||(rhs.im!=0));
  }

inline int operator&&(double lhs,complex & rhs)
  {
  return(lhs!=0)&&((rhs.re!=0)||(rhs.im!=0));
  }

inline int operator&&(complex & lhs,double rhs)
  {
  return((lhs.re!=0)||(lhs.im!=0))&&(rhs!=0);
  }

inline int operator||(complex & lhs,complex & rhs)
  {
  return((lhs.re!=0)||(lhs.im!=0))||((rhs.re!=0)||(rhs.im!=0));
  }

inline int operator||(double lhs,complex & rhs)
  {
  return(lhs!=0)||((rhs.re!=0)||(rhs.im!=0));
  }

inline int operator||(complex & lhs,double rhs)
  {
  return((lhs.re!=0)||(lhs.im!=0))||(rhs!=0);
  }

inline int operator !=(complex& lhs,complex& rhs)
  {
  return(lhs.re!=rhs.re)||(lhs.im!=rhs.im);
  }

inline int operator !=(double lhs,complex& rhs)
  {
  return(lhs!=rhs.re)||(0!=rhs.im);
  }

inline int operator !=(complex& lhs,double rhs)
  {
  return(lhs.re!=rhs)||(lhs.im!=0);
  }

inline int operator ==(complex& lhs,complex& rhs)
  {
  return(lhs.re==rhs.re)&&(lhs.im==rhs.im);
  }

inline int operator ==(double lhs,complex& rhs)
  {
  return(lhs==rhs.re)&&(0==rhs.im);
  }

inline int operator ==(complex& lhs,double rhs)
  {
  return(lhs.re==rhs)&&(lhs.im==0);
  }

inline double arg(complex& z)
  {
  return atan2(z.im,z.re);
  }

inline complex atan(complex& z)
  {
  return (log(1.-I*z)-log(1.+I*z))*I/2.;
  }

inline complex atanh(complex& z)
  {
  complex iz(-z.im,z.re);
  complex ix=atan(iz);
  return complex(ix.im,-ix.re);
  }

inline complex asinh(complex& z)
  {
  complex iz(-z.im,z.re);
  complex ix=asin(iz);
  return complex(ix.im,-ix.re);
  }

inline double norm(complex& z)
  {
  return z.re*z.re+z.im*z.im;
  }

inline double abs(complex& z)
  {
  return hypot(z.re,z.im);
  }

inline double modulus(complex& z)
  {
  return abs(z);
  }

inline complex sqrt(complex& z)
  {
  double r,theta;
  complex s;
  r=modulus(z);
  theta=arg(z);
  s.re=sqrt(r)*cos(theta/2.);
  s.im=sqrt(r)*sin(theta/2.);
  return s;
  }

inline complex pow(complex& z,int i)
  {
  complex s;
  s.re=1.;
  s.im=0.;
  while(i>0)
    {
    s*=z;
    i--;
    }
  while(i<0)
    {
    s/=z;
    i--;
    }
  return s;
  }

inline complex pow(complex& z,double x)
  {
  double r,theta;
  complex s;
  r=modulus(z);
  theta=arg(z);
  s.re=pow(r,x)*cos(theta*x);
  s.im=pow(r,x)*sin(theta*x);
  return s;
  }

inline complex pow(complex& z,complex& w)
  {
  double r,theta,x,y;
  complex s;
  r=modulus(z);
  theta=arg(z);
  x=pow(r,w.re)*exp(-theta*w.im);
  y=w.im*log(r)+w.re*theta;
  s.re=x*cos(y);
  s.im=x*sin(y);
  return s;
  }

inline complex conj(complex& z)
  {
  return complex(z.re,-z.im);
  }

inline complex cos(complex& z)
  {
  return complex(cos(z.re)*cosh(z.im),-(sin(z.re)*sinh(z.im)));
  }

inline complex acos(complex& z)
  {
  return 2.*atan(sqrt(1.-z*z)/(1.+z));
  }

inline complex cosh(complex& z)
  {
  return complex(cosh(z.re)*cos(z.im),sinh(z.re)*sin(z.im));
  }

inline complex sin(complex& z)
  {
  return complex(sin(z.re)*cosh(z.im),cos(z.re)*sinh(z.im));
  }

inline complex asin(complex& z)
  {
  return 2.*atan(z/(1.+sqrt(1.-z*z)));
  }

inline complex sinh(complex& z)
  {
  return complex(sinh(z.re)*cos(z.im),cosh(z.re)*sin(z.im));
  }

inline complex tan(complex& z)
  {
  double x=2*z.re;
  double y=2*z.im;
  double t=1.0/(cos(x)+cosh(y));
  return complex(t*sin(x),t*sinh(y));
  }

inline complex tanh(complex& z)
  {
  double x=2*z.re;
  double y=2*z.im;
  double t=1.0/(cosh(x)+cos(y));
  return complex(t*sinh(x),t*sin(y));
  }

inline complex exp(complex& z)
  {
  double x=exp(z.re);
  return complex(x*cos(z.im),x*sin(z.im));
  }

inline complex log(complex& z)
  {
  return complex(log(abs(z)),arg(z));
  }

inline complex log10(complex& z)
  {
  return complex(log10(abs(z)),arg(z));
  }

inline complex polar(double r,double theta)
  {
  return complex(r*cos(theta),r*sin(theta));
  }

complex gammal(complex z) /* ln(gamma(z)) for very large z */
  {
  return((z-0.5)*log(z)-z+log(2.*M_PI)/2.+(((-1680./z/z+1260.)/z/z-360.)/z/z+12.)/z);
  }

complex gamma(complex z)
  {
  complex g;
  g=1.;
  while(z.re<100.)
    {
    g*=z;
    z+=1.;
    }
  return(exp(gammal(z))/g);
  }

complex LambertW(complex z)
  {
  int i;
  static double eps=4E-16,em1=0.3678794411714423215955237701614608;
  complex p,e,t,w;
  if(abs(z)<DBL_EPSILON)
    return(complex(0.,0.));
  if(abs(z)<-em1+0.0001)
    {
    complex q,q2,q3,r;
    q=z+em1;
    r=sqrt(q);
    q2=q*q;
    q3=q2*q;
    return(-1.
      +2.331643981597124203363536062168*r
      -1.812187885639363490240191647568*q
      +1.936631114492359755363277457668*r*q
      -2.353551201881614516821543561516*q2
      +3.066858901050631912893148922704*r*q2
      -4.175335600258177138854984177460*q3
      +5.858023729874774148815053846119*r*q3
      -8.401032217523977370984161688514*q3*q);
    }
  if(abs(z)<1.)
    {
    p=sqrt(2.*(2.7182818284590452353602874713526625*z+1.));
    w=-1.+p*(1.+p*(-0.333333333333333333333+p*0.152777777777777777777777));
    }
  else
    w=log(z);
  if(abs(z)>3.)
    w-=log(w);
  for(i=0;i<32;i++)
    {
    e=exp(w);
    t=w*e-z;
    p=w+1.;
    t/=e*p-0.5*(p+1.)*t/p;
    w-=t;
    if(abs(t)<eps*(1.+abs(w)))
      break;
    }
  return(w);
  }

#endif
