/* a simple surface drawing program
   by Dudley J. Benton */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>
#define _WIN32_LEAN_AND_MEAN
#include <windows.h>

char*caption="push the arrow keys, Home, End, PgUp, PgDn, Tab, BkSp, Esc, spacebar";

extern int __argc;
extern char**__argv;
extern char**_environ;

HINSTANCE hInst;
HWND hMain;

struct{
  HDC dc;
  HBITMAP bm;
  BITMAPINFO bi;
  BYTE*bits;
  }buffer;

struct{int back,pen,x,y,z;}angle={0,0,0,30,0};
struct{double R,S,T;int xc,xm,xx,yc,ym,yx;}scale;

typedef struct{double X,Y,Z;}NODE;
typedef struct{NODE N[3];POINT p[3];int z;DWORD rgb;}ELEM;
ELEM*Elem;
int Me,Ne;

void Abort(int code,char*format,...)
  {
  char bufr[1024],head[32];
  va_list arg_marker;
  va_start(arg_marker,format);
  if(sprintf(head,"Fatal Error Code %i",code)>=sizeof(head))
    FatalAppExit(0,"Abort(): caption buffer overflow");
  if(vsprintf(bufr,format,arg_marker)>=sizeof(bufr))
    FatalAppExit(0,"Abort(): text buffer overflow");
  MessageBox(hMain,bufr,head,MB_APPLMODAL|MB_ICONSTOP|MB_OK);
  ExitProcess(1);
  }

void*allocate(int line,size_t count,size_t size)
  {
  void*ptr;
  if((ptr=(void*)calloc(count,size))==NULL)
    Abort(line,"can't allocate memory %ux%u=%u bytes",count,size,count*size);
  return(ptr);
  }

void*reallocate(int line,void*old_ptr,size_t old_count,size_t new_count,size_t size)
  {
  void*new_ptr;
  if((new_ptr=calloc(new_count,size))==NULL)
    Abort(line,"can't reallocate memory %ux%u=%u -> %lux%u=%lu bytes",old_count,size,old_count*size,new_count,size,new_count*size);
  if(old_ptr&&old_count)
    {
    memcpy(new_ptr,old_ptr,min(old_count,new_count)*size);
    free(old_ptr);
    }
  return(new_ptr);
  }

int nint(double d)
  {
  if(d>0.)
    return((int)(d+0.5));
  if(d<0.)
    return((int)(d-0.5));
  return(0);
  }

double sq(double x)
  {
  return(x*x);
  }

void TransformObjects()
  {
  int k,n,xi,xn,yi,yn,zi,zn;
  int r11,r12,r13,r21,r22,r23,r31,r32,r33;
  double C1,C2,C3,S1,S2,S3;
  RECT rc;
  GetClientRect(hMain,&rc);
  if(scale.S<=0.)
    {
    double Xm,Xx,Ym,Yx,Zm,Zx;
    Xm=Ym=Zm=DBL_MAX;
    Xx=Yx=Zx=-DBL_MAX;
    for(n=0;n<Ne;n++)
      {
      for(k=0;k<3;k++)
        {
        Xm=min(Xm,Elem[n].N[k].X);
        Xx=max(Xx,Elem[n].N[k].X);
        Ym=min(Ym,Elem[n].N[k].Y);
        Yx=max(Yx,Elem[n].N[k].Y);
        Zm=min(Zm,Elem[n].N[k].Z);
        Zx=max(Zx,Elem[n].N[k].Z);
        }
      }
    scale.S=rc.right/sqrt(sq(Xx-Xm)+sq(Yx-Ym)+sq(Zx-Zm));
    scale.R=scale.T=1.;
    }
  S1=sin(M_PI*angle.x/180.);          /* direction cosines */
  S2=sin(M_PI*angle.y/180.);
  S3=sin(M_PI*angle.z/180.);
  C1=cos(M_PI*angle.x/180.);
  C2=cos(M_PI*angle.y/180.);
  C3=cos(M_PI*angle.z/180.);
  r11=nint((C1*C3-S1*S2*S3)*32767);   /* rotation matrix */
  r12=nint((S1*C3+C1*S2*S3)*32767);
  r13=nint(-C2*S3*32767);
  r21=nint(-S1*C2*32767);
  r22=nint(C1*C2*32767);
  r23=nint(S2*32767);
  r31=nint((C1*S3+S1*S2*C3)*32767);
  r32=nint((S1*S3-C1*S2*C3)*32767);
  r33=nint((C2*C3)*32767);
  scale.xm=scale.ym=32767;
  scale.xx=scale.yx=-32767;
  for(n=0;n<Ne;n++)
    {
    Elem[n].z=0;
    for(k=0;k<3;k++)
      {
      xi=nint(Elem[n].N[k].X*scale.S*scale.R);
      yi=nint(Elem[n].N[k].Y*scale.S*scale.T);
      zi=nint(Elem[n].N[k].Z*scale.S*scale.R);
      xn=(r11*xi+r12*yi+r13*zi)/32767;
      yn=(r21*xi+r22*yi+r23*zi)/32767;
      zn=(r31*xi+r32*yi+r33*zi)/32767;
      Elem[n].p[k].x=xn;
      Elem[n].p[k].y=yn;
      Elem[n].z+=zn;
      scale.xm=min(scale.xm,xn);
      scale.xx=max(scale.xx,xn);
      scale.ym=min(scale.ym,yn);
      scale.yx=max(scale.yx,yn);
      }
    Elem[n].z/=3;
    }
  scale.xc=(scale.xm+scale.xx)/2;
  scale.yc=(scale.ym+scale.yx)/2;
  scale.xc-=rc.right/2;         /* adjust new center */
  scale.yc-=rc.bottom/2;
  for(n=0;n<Ne;n++)             /* apply shift to center */
    {
    for(k=0;k<3;k++)
      {
      Elem[n].p[k].x-=scale.xc;
      Elem[n].p[k].y-=scale.yc;
      }
    }
  }

int CompareDepth(const void*v1,const void*v2)
  {
  return(Elem[*(int*)v1].z-Elem[*(int*)v2].z);
  }

void RenderObjects()
  {
  char text[1024];
  int i,n,*depth;
  HBRUSH hbr;
  RECT rc;
  GetClientRect(hMain,&rc);
  if(angle.back%2)
    FillRect(buffer.dc,&rc,GetStockObject(WHITE_BRUSH));
  else
    FillRect(buffer.dc,&rc,GetStockObject(BLACK_BRUSH));
  TransformObjects();
  depth=allocate(__LINE__,Ne,sizeof(int));
  for(n=0;n<Ne;n++)
    depth[n]=n;
  qsort(depth,Ne,sizeof(int),CompareDepth);
  if(angle.pen%3==0)
    SelectObject(buffer.dc,GetStockObject(BLACK_PEN));
  else if(angle.pen%3==1)
    SelectObject(buffer.dc,GetStockObject(WHITE_PEN));
  else
    SelectObject(buffer.dc,GetStockObject(NULL_PEN));
  for(n=0;n<Ne;n++)
    {
    i=depth[n];
    hbr=CreateSolidBrush(Elem[i].rgb);
    hbr=SelectObject(buffer.dc,hbr);
    Polygon(buffer.dc,Elem[i].p,3);
    hbr=SelectObject(buffer.dc,hbr);
    DeleteObject(hbr);
    }
  free(depth);
  RedrawWindow(hMain,NULL,NULL,RDW_INVALIDATE|RDW_ERASE);
  sprintf(text,"%s %i:%i:%i %.2lf:%.2lf",caption,angle.x,angle.y,angle.z,scale.R,scale.T);
  SetWindowText(hMain,text);
  }

void AddTriangle(double X1,double Y1,double Z1,double X2,double Y2,double Z2,double X3,double Y3,double Z3,DWORD rgb)
  {
  if(Me==0)
    {
    Me=1024;
    Elem=allocate(__LINE__,Me,sizeof(ELEM));
    }
  else if(Ne>=Me)
    {
    Me+=1024;
    Elem=reallocate(__LINE__,Elem,Ne,Me,sizeof(ELEM));
    }
  Elem[Ne].N[0].X=X1;
  Elem[Ne].N[0].Y=Y1;
  Elem[Ne].N[0].Z=Z1;
  Elem[Ne].N[1].X=X2;
  Elem[Ne].N[1].Y=Y2;
  Elem[Ne].N[1].Z=Z2;
  Elem[Ne].N[2].X=X3;
  Elem[Ne].N[2].Y=Y3;
  Elem[Ne].N[2].Z=Z3;
  Elem[Ne++].rgb=rgb;
  }

DWORD Rainbow[]={   0xFF0000,0xFF0800,0xFF1000,0xFF1800,0xFF2000,0xFF2800,
  0xFF3000,0xFF3800,0xFF4000,0xFF4800,0xFF5000,0xFF5800,0xFF6000,0xFF6800,
  0xFF7000,0xFF7800,0xFF8000,0xFF8800,0xFF9000,0xFF9800,0xFFA000,0xFFA800,
  0xFFB000,0xFFB800,0xFFC000,0xFFC800,0xFFD000,0xFFD800,0xFFE000,0xFFE800,
  0xFFF000,0xFFF800,0xFFFF00,0xF8FF00,0xF0FF00,0xE8FF00,0xE0FF00,0xD8FF00,
  0xD0FF00,0xC8FF00,0xC0FF00,0xB8FF00,0xB0FF00,0xA8FF00,0xA0FF00,0x98FF00,
  0x90FF00,0x88FF00,0x80FF00,0x78FF00,0x70FF00,0x68FF00,0x60FF00,0x58FF00,
  0x50FF00,0x48FF00,0x40FF00,0x38FF00,0x30FF00,0x28FF00,0x20FF00,0x18FF00,
  0x10FF00,0x08FF00,0x00FF00,0x00FF08,0x00FF10,0x00FF18,0x00FF20,0x00FF28,
  0x00FF30,0x00FF38,0x00FF40,0x00FF48,0x00FF50,0x00FF58,0x00FF60,0x00FF68,
  0x00FF70,0x00FF78,0x00FF80,0x00FF88,0x00FF90,0x00FF98,0x00FFA0,0x00FFA8,
  0x00FFB0,0x00FFB8,0x00FFC0,0x00FFC8,0x00FFD0,0x00FFD8,0x00FFE0,0x00FFE8,
  0x00FFF0,0x00FFF8,0x00FFFF,0x00F8FF,0x00F0FF,0x00E8FF,0x00E0FF,0x00D8FF,
  0x00D0FF,0x00C8FF,0x00C0FF,0x00B8FF,0x00B0FF,0x00A8FF,0x00A0FF,0x0098FF,
  0x0090FF,0x0088FF,0x0080FF,0x0078FF,0x0070FF,0x0068FF,0x0060FF,0x0058FF,
  0x0050FF,0x0048FF,0x0040FF,0x0038FF,0x0030FF,0x0028FF,0x0020FF,0x0018FF,
  0x0010FF,0x0008FF,0x0000FF,0x0800FF,0x1000FF,0x1800FF,0x2000FF,0x2800FF,
  0x3000FF,0x3800FF,0x4000FF,0x4800FF,0x5000FF,0x5800FF,0x6000FF,0x6800FF,
  0x7000FF,0x7800FF,0x8000FF,0x8800FF,0x9000FF,0x9800FF,0xA000FF,0xA800FF,
  0xB000FF,0xB800FF,0xC000FF,0xC800FF,0xD000FF,0xD800FF,0xE000FF,0xE800FF,
  0xF000FF,0xF800FF,0xFF00FF};
int colors=sizeof(Rainbow)/sizeof(Rainbow[0]);

void LoadSurface(char*name)
  {
  char bufr[128];
  int color,l,nx,ny,nz,x,y,z;
  double R,S,*X,X1,X2,X3,X4,Xc,*Y,Y1,Y2,Y3,Y4,Yc,*Z,Z1,Z2,Z3,Z4,Zc,Zm,Zx;
  FILE*fp;
  if((fp=fopen(name,"rt"))==NULL)
    Abort(__LINE__,"can't open file %s",name);
  l=0;
  if(fgets(bufr,sizeof(bufr),fp))
    l++;
  else
    Abort(__LINE__,"unexpected end of file at line %i",l);
  if(sscanf(bufr,"%i",&nx)!=1)
    Abort(__LINE__,"scan error on line %i",l);
  if(nx<2)
    Abort(__LINE__,"Nx<2");
  if(nx>8096)
    Abort(__LINE__,"Nx>8096");
  X=allocate(__LINE__,nx,sizeof(double));
  for(x=0;x<nx;x++)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      Abort(__LINE__,"unexpected end of file at line %i",l);
    if(sscanf(bufr,"%lf",X+x)!=1)
      Abort(__LINE__,"scan error on line %i",l);
    if(x>0)
      if(X[x]<=X[x-1])
        Abort(__LINE__,"X values out of order at line %i",l);
     }
  if(fgets(bufr,sizeof(bufr),fp))
    l++;
  else
    Abort(__LINE__,"unexpected end of file at line %i",l);
  if(sscanf(bufr,"%i",&ny)!=1)
    Abort(__LINE__,"scan error on line %i",l);
  if(ny<2)
    Abort(__LINE__,"Ny<2");
  if(ny>8096)
    Abort(__LINE__,"Ny>8096");
  Y=allocate(__LINE__,ny,sizeof(double));
  for(y=0;y<ny;y++)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      Abort(__LINE__,"unexpected end of file at line %i",l);
    if(sscanf(bufr,"%lf",Y+y)!=1)
      Abort(__LINE__,"scan error on line %i",l);
    if(y>0)
      if(Y[y]<=Y[y-1])
        Abort(__LINE__,"Y values out of order at line %i",l);
     }
  if(fgets(bufr,sizeof(bufr),fp))
    l++;
  else
    Abort(__LINE__,"unexpected end of file at line %i",l);
  if(sscanf(bufr,"%i",&nz)!=1)
    Abort(__LINE__,"scan error on line %i",l);
  if(nz!=nx*ny)
    Abort(__LINE__,"Nz<>Nx*Ny");
  Z=allocate(__LINE__,nz,sizeof(double));
  Zm=DBL_MAX;
  Zx=-DBL_MAX;
  for(z=0;z<nz;z++)
    {
    if(fgets(bufr,sizeof(bufr),fp))
      l++;
    else
      Abort(__LINE__,"unexpected end of file at line %i",l);
    if(sscanf(bufr,"%lf",Z+z)!=1)
      Abort(__LINE__,"scan error on line %i",l);
    Zm=min(Zm,Z[z]);
    Zx=max(Zx,Z[z]);
    }
  fclose(fp);
  R=hypot(X[nx-1]-X[0],Y[ny-1]-Y[0]);
  Xc=(X[nx-1]+X[0])/2.;
  Yc=(Y[ny-1]+Y[0])/2.;
  S=3.*(Zx-Zm);
  for(x=1;x<nx;x++)
    {
    X1=X4=X[x-1];
    X2=X3=X[x];
    for(y=1;y<ny;y++)
      {
      Y1=Y2=Y[y-1];
      Y3=Y4=Y[y];
      Z1=Z[nx*(y-1)+x-1];
      Z2=Z[nx*(y-1)+x  ];
      Z3=Z[nx* y   +x  ];
      Z4=Z[nx* y   +x-1];
      Zc=(Z1+Z2+Z3)/3.;
      color=nint((Zc-Zm)*colors/(Zx-Zm));
      color=max(0,min(colors-1,color));
      AddTriangle((X1-Xc)/R,(Zx-Z1)/S,(Yc-Y1)/R,
                  (X2-Xc)/R,(Zx-Z2)/S,(Yc-Y2)/R,
                  (X3-Xc)/R,(Zx-Z3)/S,(Yc-Y3)/R,
                  Rainbow[color]);
      Zc=(Z1+Z3+Z4)/3.;
      color=nint((Zc-Zm)*colors/(Zx-Zm));
      color=max(0,min(colors-1,color));
      AddTriangle((X1-Xc)/R,(Zx-Z1)/S,(Yc-Y1)/R,
                  (X3-Xc)/R,(Zx-Z3)/S,(Yc-Y3)/R,
                  (X4-Xc)/R,(Zx-Z4)/S,(Yc-Y4)/R,
                  Rainbow[color]);
      }
    }
  free(X);
  free(Y);
  free(Z);
  scale.S=0.;
  }

LRESULT WINAPI MainProc(HWND hWnd,DWORD wMsg,DWORD wParam,LPARAM lParam)
  {
  if(wMsg==WM_CLOSE)
    PostQuitMessage(0);
  if(wMsg==WM_COMMAND)
    return(FALSE);
  if(wMsg==WM_CREATE)
    return(FALSE);
  if(wMsg==WM_KEYDOWN)
    {
    if(LOWORD(wParam)==VK_RIGHT)
      angle.x+=5;
    else if(LOWORD(wParam)==VK_LEFT)
      angle.x-=5;
    else if(LOWORD(wParam)==VK_DOWN)
      angle.y+=5;
    else if(LOWORD(wParam)==VK_UP)
      angle.y-=5;
    else if(LOWORD(wParam)==VK_END)
      angle.z+=5;
    else if(LOWORD(wParam)==VK_HOME)
      angle.z-=5;
    else if(LOWORD(wParam)==VK_PRIOR)
      scale.R*=1.05;
    else if(LOWORD(wParam)==VK_NEXT)
      scale.R/=1.05;
    else if(LOWORD(wParam)==VK_BACK)
      scale.T*=1.05;
    else if(LOWORD(wParam)==VK_TAB)
      scale.T/=1.05;
    else if(LOWORD(wParam)==VK_SPACE)
      angle.pen++;
    else if(LOWORD(wParam)==VK_ESCAPE)
      angle.back++;
    else
      return(FALSE);
    RenderObjects();
    return(FALSE);
    }
  if(wMsg==WM_PAINT)
    {
    HDC hDC;
    PAINTSTRUCT ps;
    RECT rc;
    hDC=BeginPaint(hWnd,&ps);
    GetClientRect(hWnd,&rc);
    BitBlt(hDC,0,0,rc.right,rc.bottom,buffer.dc,0,0,SRCCOPY);
    EndPaint(hWnd,&ps);
    return(FALSE);
    }
  return(DefWindowProc(hWnd,wMsg,wParam,lParam));
  }

void CreateWindows()
  {
  int h,high,w,wide;
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=7*wide/8;
  h=7*high/8;
  if((hMain=CreateWindow("MAIN",caption,
      WS_OVERLAPPEDWINDOW|WS_CAPTION|WS_SYSMENU|WS_VISIBLE,
      (wide-w)/2,(high-h)/2,w,h,NULL,NULL,hInst,NULL))==NULL)
    Abort(__LINE__,"can't create window\nerror code %i",GetLastError());
  }

void Register(char*cName,void*wProc,HBRUSH hBrush,DWORD dCursor)
  {
  WNDCLASS wc;
  memset(&wc,0,sizeof(WNDCLASS));
  wc.hInstance    =hInst;
  wc.style        =CS_HREDRAW|CS_VREDRAW;
  wc.lpszClassName=cName;
  wc.lpfnWndProc  =(WNDPROC)wProc;
  wc.hbrBackground=hBrush;
  wc.hCursor      =LoadCursor(NULL,(char*)dCursor);
  if(!RegisterClass(&wc))
    Abort(__LINE__,"can't register class\nerror code %i",GetLastError());
  }

void RegisterClasses()
  {
  Register("MAIN",MainProc,GetStockObject(NULL_BRUSH),(DWORD)IDC_ARROW);
  }

void CreateBuffer()
  {
  int high,w,wide;
  buffer.dc=CreateCompatibleDC(NULL);
  wide=GetSystemMetrics(SM_CXSCREEN);
  high=GetSystemMetrics(SM_CYSCREEN);
  w=3*wide;
  while(w%4)
    w++;
  memset(&buffer.bi,0,sizeof(buffer.bi));
  buffer.bi.bmiHeader.biSize=sizeof(BITMAPINFOHEADER);
  buffer.bi.bmiHeader.biPlanes=1;
  buffer.bi.bmiHeader.biBitCount=24;
  buffer.bi.bmiHeader.biWidth=wide;
  buffer.bi.bmiHeader.biHeight=high;
  buffer.bi.bmiHeader.biSizeImage=w*high;
  buffer.dc=CreateCompatibleDC(NULL);
  buffer.bm=CreateDIBSection(buffer.dc,&buffer.bi,DIB_RGB_COLORS,&buffer.bits,NULL,0);
  SelectObject(buffer.dc,buffer.bm);
  }

int WINAPI WinMain(HINSTANCE hCurrent,HINSTANCE hPrevious,char*lCommand,int nShow)
  {
  char*fname;
  MSG wm;
  hInst=hCurrent;
  CreateBuffer();
  if(__argc>1)
    fname=__argv[1];
  else
    {
    HANDLE hf;
    WIN32_FIND_DATA wf;
    memset(&wf,0,sizeof(wf));
    hf=FindFirstFile("*.tb2",&wf);
    if(hf==NULL||hf==(HANDLE)INVALID_HANDLE_VALUE)
      Abort(__LINE__,"can't find any file *.tb2");
    fname=wf.cFileName;
    }
  LoadSurface(fname);
  RegisterClasses();
  CreateWindows();
  RenderObjects();
  while(GetMessage(&wm,NULL,0,0))
    {
    TranslateMessage(&wm);
    DispatchMessage(&wm);
    }
  return((int)wm.wParam);
  }
